/* eslint-disable @typescript-eslint/no-unsafe-return */
/* eslint-disable @typescript-eslint/member-delimiter-style */
/* eslint-disable prefer-arrow/prefer-arrow-functions */
import * as fs from 'fs';
import * as path from 'path';

import { relativeDate } from './datetime';
import { getLogger } from './loggers';
import {
  getPostOwner, sendPost, ScreenNameNormalizer as normalizer,
  isValidUrlSegment, linkBuilder, parseLink, urlSegmentToId
} from './twitter';
import { BigNumOps } from './utils';

const logger = getLogger('command');

function parseCmd(message: string): {
  cmd: string;
  args: string[];
} {
  message = message.trim();
  message = message.replace('\\\\', '\\0x5c');
  message = message.replace('\\\"', '\\0x22');
  message = message.replace('\\\'', '\\0x27');
  const strs = message.match(/'[\s\S]*?'|(?:\S+=)?"[\s\S]*?"|\S+/mg);
  const cmd = strs?.length ? strs[0].length ? strs[0].substring(0, 1) === '/' ? strs[0].substring(1) : '' : '' : '';
  const args = (strs ?? []).slice(1).map(arg => {
    arg = arg.replace(/^(\S+=)?["']+(?!.*=)|["']+$/g, '$1');
    arg = arg.replace('\\0x27', '\\\'');
    arg = arg.replace('\\0x22', '\\\"');
    arg = arg.replace('\\0x5c', '\\\\');
    return arg;
  });
  return {
    cmd,
    args,
  };
}

function linkFinder(userName: string, chat: IChat, lock: ILock): [string, number] {
  const normalizedLink = linkBuilder({userName});
  const link = Object.keys(lock.threads).find(realLink =>
    normalizedLink === realLink.replace(/\/@/, '/').toLowerCase()
  );
  if (!link) return [null, -1];
  const index = lock.threads[link].subscribers.findIndex(({chatID, chatType}) =>
    chat.chatID === chatID && chat.chatType === chatType
  );
  return [link, index];
}

function sub(chat: IChat, args: string[], reply: (msg: string) => any,
  lock: ILock, lockfile: string
): void {
  if (chat.chatType === ChatType.Temp) {
    return reply('请先添加机器人为好友。');
  }
  if (args.length === 0) {
    return reply('找不到要订阅的链接。');
  }
  const matched = parseLink(args[0]);
  if (!matched) {
    return reply(`订阅链接格式错误：
示例：
https://www.instagram.com/tomoyo_kurosawa_/
https://www.instagram.com/p/B6GHRSmgV-7/`);
  }
  let offset = '0';
  const subscribeTo = (link: string, config: {id?: number, msg?: string} = {}) => {
    const {id, msg = `已为此聊天订阅 ${link}`} = config;
    if (id) {
      lock.feed.push(link);
      lock.threads[link] = {
        id,
        offset,
        subscribers: [],
        updatedAt: '',
      };
    }
    lock.threads[link].subscribers.push(chat);
    logger.warn(`chat ${JSON.stringify(chat)} has subscribed ${link}`);
    fs.writeFileSync(path.resolve(lockfile), JSON.stringify(lock));
    reply(msg);
  };
  const tryFindSub = (userName: string) => {
    const [realLink, index] = linkFinder(userName, chat, lock);
    if (index > -1) { reply('此聊天已订阅此链接。'); return true; }
    if (realLink) { subscribeTo(realLink); return true; }
    return false;
  };
  const newSub = (userName: string) => {
    const link = linkBuilder(matched);
    let msg: string;
    if (offset !== '0') {
      msg = `已为此聊天订阅 ${link} 并回溯到 ${args[0].replace(/\?.*/, '')}（含）之后的第一条动态`;
    }
    return subscribeTo(link, {id: Number(userName.split(':')[1]), msg});
  };
  if (matched.postUrlSegment) {
    offset = BigNumOps.plus(urlSegmentToId(matched.postUrlSegment), '-1');
    getPostOwner(matched.postUrlSegment).then(userName => {
      delete matched.postUrlSegment;
      matched.userName = userName.split(':')[0];
      if (!tryFindSub(matched.userName)) newSub(userName);
    }).catch((parsedErr: Error) => {
      reply(parsedErr.message);
    });
  } else if (!tryFindSub(matched.userName)) {
    normalizer.normalizeLive(matched.userName).then(userName => {
      if (!userName) return reply(`找不到用户 ${matched.userName.replace(/^@?(.*)$/, '@$1')}。`);
      else newSub(userName);
    });
  }
}

function unsub(chat: IChat, args: string[], reply: (msg: string) => any,
  lock: ILock, lockfile: string
): void {
  if (chat.chatType === ChatType.Temp) {
    return reply('请先添加机器人为好友。');
  }
  if (args.length === 0) {
    return reply('找不到要退订的链接。');
  }
  const match = parseLink(args[0])?.userName;
  if (!match) {
    return reply('链接格式有误。');
  }
  const [link, index] = linkFinder(match, chat, lock);
  if (index === -1) return list(chat, args, msg => reply('您没有订阅此链接。\n' + msg), lock);
  else {
    lock.threads[link].subscribers.splice(index, 1);
    fs.writeFileSync(path.resolve(lockfile), JSON.stringify(lock));
    logger.warn(`chat ${JSON.stringify(chat)} has unsubscribed ${link}`);
    return reply(`已为此聊天退订 ${link}`);
  }
}

function list(chat: IChat, _: string[], reply: (msg: string) => any, lock: ILock): void {
  if (chat.chatType === ChatType.Temp) {
    return reply('请先添加机器人为好友。');
  }
  const links = [];
  Object.keys(lock.threads).forEach(key => {
    if (lock.threads[key].subscribers.find(({chatID, chatType}) => 
      chat.chatID === chatID && chat.chatType === chatType
    )) links.push(`${key} ${relativeDate(lock.threads[key].updatedAt)}`);
  });
  return reply('此聊天中订阅的 Instagram 动态链接：\n' + links.join('\n'));
}

function view(chat: IChat, args: string[], reply: (msg: string) => any): void {
  if (args.length === 0) {
    return reply('找不到要查看的链接。');
  }
  const match = isValidUrlSegment(args[0]) && args[0] || parseLink(args[0])?.postUrlSegment;
  if (!match) {
    return reply('链接格式有误。');
  }
  try {
    sendPost(match, chat);
  } catch (e) {
    reply('机器人尚未加载完毕，请稍后重试。');
  }
}

export { parseCmd, sub, list, unsub, view };
